from typing import Any, Callable, Generator, Union

import bson


class PyObjectId(bson.ObjectId):
    @classmethod
    def __get_validators__(cls) -> Generator[Callable[[Any], bson.ObjectId], None, None]:
        yield cls.validate

    @classmethod
    def validate(cls, v: Union[bson.ObjectId, bytes]) -> bson.ObjectId:
        if not bson.ObjectId.is_valid(v):
            raise ValueError("Invalid object id")
        return bson.ObjectId(v)

    @classmethod
    def __modify_schema__(cls, field_schema: dict[str, Any]) -> None:
        field_schema.update(type="string")



from __future__ import annotations

from typing import Any, Optional

from .base import Operator
from .comparison import ComparisonOperator


class Expr(Operator):
    def __init__(self, op: Operator):
        self.op = op

    def expression(self) -> dict[str, Any]:
        return {"$expr": self.op.expression()}


class AEqual(ComparisonOperator):
    def __init__(self, field: str, value: Optional[Any] = None):
        super().__init__(field=field, value=value)

    def expression(self) -> dict[str, Any]:
        return {"$eq": [self.field, self.value]}



from __future__ import annotations

import datetime
from collections.abc import Callable, Generator
from typing import Any, Union

import pendulum
from pydantic.datetime_parse import parse_datetime


class DateTime(pendulum.DateTime):
    @classmethod
    def __get_validators__(cls) -> Generator[Callable[[Any], pendulum.DateTime], None, None]:
        yield parse_datetime  # type: ignore
        yield cls._instance

    @classmethod
    def _instance(cls, v: Union[datetime.datetime, pendulum.DateTime]) -> pendulum.DateTime:
        return pendulum.instance(dt=v, tz=pendulum.UTC)



from typing import Any

from .base import Operator
from .stage import Stage


class Pipeline(Operator):
    def __init__(self, *stages: Stage):
        self.stages = list(stages)

    def expression(self) -> list[dict[str, Any]]:
        return [expression for stage in self.stages if (expression := stage.expression())]



from . import command, dependency, entity, event, field, query, repository, schema, types
from .container import BaseModel, TimeStampedModel

__all__ = [
    "BaseModel",
    "TimeStampedModel",
    "entity",
    "field",
    "schema",
    "repository",
    "query",
    "types",
    "command",
    "event",
    "dependency",
]



from typing import Optional

from .base import Operator
from .types import DictExpression, ListExpression


class Pagination(Operator):
    def __init__(
        self,
        page: Optional[int] = None,
        per_page: Optional[int] = None,
        metadata_ops: Optional[list[Operator]] = None,
    ):
        if not ((page is not None and per_page is not None) or (page is None and per_page is None)):
            raise ValueError("`page` and `per_page` are mutual inclusive")

        self.page = page
        self.per_page = per_page
        self.metadata_ops = metadata_ops or []

        self.can_paginate = self.page is not None and self.per_page is not None

    def expression(self) -> DictExpression:
        return {
            "$facet": {
                "metadata": [{"$count": "total"}, *[op.expression() for op in self.metadata_ops]],
                "items": self._pagination_expression(),
            }
        }

    def _pagination_expression(self) -> ListExpression:
        if not self.can_paginate:
            return []

        return [
            {"$skip": (self.page - 1) * self.per_page},  # type: ignore
            {"$limit": self.per_page},
        ]



from collections.abc import Callable
from typing import Any, Optional

import pendulum
from pydantic import Field

from . import types


class _IdField:
    def __call__(self, alias: Optional[str] = None, **kwargs) -> Any:
        return Field(alias=alias or "_id", **kwargs)  # type: ignore


class _DatetimeField:
    def __call__(self, **kwargs) -> types.DateTime:
        return Field(default_factory=pendulum.DateTime.utcnow, **kwargs)  # type: ignore


IdField: Callable[..., Any] = _IdField()
DateTimeField: Callable[..., types.DateTime] = _DatetimeField()



from __future__ import annotations

import abc
import functools
import operator
from typing import Any


class Operator:
    @abc.abstractmethod
    def expression(self) -> Any:
        raise NotImplementedError


class Merge(Operator):
    def __init__(self, *ops: Operator, initial: Any):
        self.ops = list(ops)
        self.initial = initial

    def expression(self) -> Any:
        return functools.reduce(operator.or_, [op.expression() for op in self.ops], self.initial)

    @classmethod
    def dict(cls, *ops: Operator) -> Merge:
        return cls(*ops, initial={})



import functools
import warnings
from collections.abc import Callable
from typing import Any


def deprecated(message: str, stack_level: int = 1):
    def wrapper(func: Callable[..., Any]):
        @functools.wraps(func)
        def inner(*args, **kwargs):
            warnings.warn(message, DeprecationWarning, stacklevel=stack_level)
            return func(*args, **kwargs)

        return inner

    return wrapper



from __future__ import annotations

from typing import Any

import inflection

DictExpression = dict[str, Any]
ListExpression = list[DictExpression]


class MongoKeyword(str):
    def __new__(cls, v: str):
        if v != _camelize(v):
            raise ValueError(f"Cannot interpret {v} as valid mongo keyword")

        return super().__new__(cls, v)

    @classmethod
    def from_py(cls, v: str) -> MongoKeyword:
        return cls(_camelize(v))


def _camelize(v: str) -> str:
    return inflection.camelize(v.strip("_"), uppercase_first_letter=False)



import functools
import inspect
from collections.abc import Iterable
from typing import Any, Optional, Protocol

from dependency_injector import containers, providers

from compose.utils import deprecated


class Wirer(Protocol):
    def __call__(
        self,
        container: containers.Container,
        modules: Optional[Iterable[str]] = None,
        from_package: Optional[str] = None,
    ) -> None:
        ...


def create_wirer(packages: Iterable[str]) -> Wirer:
    def wire_container(
        container: containers.Container,
        modules: Optional[Iterable[str]] = None,
        from_package: Optional[str] = None,
    ) -> None:
        container.check_dependencies()
        container.wire(modules=modules, packages=packages, from_package=from_package)

    return wire_container


@functools.lru_cache(32)
def resolve(type_: type[Any], container_cls: type[containers.Container]) -> providers.Factory:
    """
    의존성 전체 등록 경로를 참조하지 않고 의존성을 해결합니다. 다른 패키지의 의존성을 참조하는 경우
    의존 대상 선언 경로에 깊게 의존하는 것을 방지합니다. `container_cls`는 최상위 컨테이너일수도,
    의존성이 등록된 (하위) 컨테이너일수도 있습니다. 클래스 대상으로만 작동합니다.
    """
    if not inspect.isclass(type_):
        raise ValueError("Only class can be resolved")

    for provider in container_cls.traverse([providers.Factory]):
        provider_cls = provider.cls
        if not (inspect.isclass(provider_cls) or inspect.ismethod(provider_cls)):
            continue

        cls = provider_cls.__self__ if inspect.ismethod(provider_cls) else provider_cls
        if cls.__name__ == type_.__name__:  # type: ignore
            return provider

    raise ValueError(f"Cannot find {type_.__name__} from given container")


resolve_dependency = deprecated(
    "`resolve_dependency` is deprecated and will be removed in a future version. "
    "Use `resolve` instead."
)(resolve)



from __future__ import annotations

import abc
from typing import Any, Optional

from .base import Merge, Operator
from .logical import And, LogicalOperator, Or
from .types import DictExpression, ListExpression, MongoKeyword


class Stage(Operator):
    @abc.abstractmethod
    def expression(self) -> DictExpression:
        raise NotImplementedError


class Match(Stage):
    def __init__(self, op: LogicalOperator):
        self.op = op

    def expression(self) -> DictExpression:
        if not (expression := self.op.expression()):
            return {}
        return {"$match": expression}

    @classmethod
    def and_(cls, *ops: Operator) -> Match:
        return cls(And(*ops))

    @classmethod
    def or_(cls, *ops: Operator) -> Match:
        return cls(Or(*ops))


class Sort(Stage):
    def __init__(self, *ops: Operator):
        self.ops = list(ops)

    def expression(self) -> DictExpression:
        if not (merged := Merge.dict(*self.ops).expression()):
            raise ValueError("Expression cannot be empty")
        return {"$sort": merged}


class Specification(Operator):
    def __init__(self, field: str, spec: Any):
        self.field = field
        self.spec = spec

    def expression(self) -> DictExpression:
        return {self.field: self.spec}


class Project(Stage):
    def __init__(self, *specs: Specification):
        self.specs = list(specs)

    def expression(self) -> DictExpression:
        return {"$project": Merge.dict(*self.specs).expression()}


class Lookup(Stage):
    def __init__(self, from_: str, as_: str):
        self.from_ = from_
        self.as_ = as_

    def expression(self) -> DictExpression:
        return {
            "$lookup": {
                MongoKeyword.from_py(field): value for field, value in self.__dict__.items()
            }
        }


class MatchLookup(Lookup):
    def __init__(self, from_: str, as_: str, local_field: str, foreign_field: str):
        super().__init__(from_=from_, as_=as_)
        self.local_field = local_field
        self.foreign_field = foreign_field


class SubQueryLookup(Lookup):
    def __init__(self, from_: str, as_: str, let: str, pipeline: ListExpression):
        super().__init__(from_=from_, as_=as_)
        self.let = let
        self.pipeline = pipeline


class Unwind(Stage):
    def __init__(
        self,
        path: str,
        include_array_index: Optional[str] = None,
        preserve_null_and_empty_arrays: Optional[bool] = None,
    ):
        self.path = path
        self.include_array_index = include_array_index
        self.preserve_null_and_empty_arrays = preserve_null_and_empty_arrays

    def expression(self) -> DictExpression:
        return {
            "$unwind": {
                MongoKeyword.from_py(field): value for field, value in self.__dict__.items()
            }
        }


class Set(Stage):
    def __init__(self, *specs: Specification):
        self.specs = list(specs)

    def expression(self) -> DictExpression:
        return {"$set": Merge.dict(*self.specs).expression()}



from .base import Operator


class SortBy(Operator):
    def __init__(self, field: str, direction: int):
        self.field = field
        self.direction = direction

    def expression(self) -> dict[str, int]:
        return {self.field: self.direction}



from __future__ import annotations

import json
from collections.abc import Callable
from typing import TYPE_CHECKING, Any, Optional, Union

import bson
from pydantic import BaseModel as PydanticBaseModel

from . import field, types

if TYPE_CHECKING:
    from pydantic.typing import AbstractSetIntStr, DictStrAny, MappingIntStrAny


class BaseModel(PydanticBaseModel):
    def copy(
        self,
        *,
        include: Optional[Union[AbstractSetIntStr, MappingIntStrAny]] = None,
        exclude: Optional[Union[AbstractSetIntStr, MappingIntStrAny]] = None,
        update: Optional[DictStrAny] = None,
        deep: bool = False,
        strict: bool = True,
    ) -> BaseModel:
        if strict and (diff := set((update or {}).keys()) - set(self.__fields__.keys())):
            raise AttributeError(f"{self.__class__.__name__} has no attributes: {', '.join(diff)}")

        return super().copy(
            include=include,
            exclude=exclude,
            update=update,
            deep=deep,
        )

    def encode(
        self,
        *,
        include: Optional[Union[AbstractSetIntStr, MappingIntStrAny]] = None,
        exclude: Optional[Union[AbstractSetIntStr, MappingIntStrAny]] = None,
        by_alias: bool = False,
        skip_defaults: Optional[bool] = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        encoder: Optional[Callable[[Any], Any]] = None,
        models_as_dict: bool = True,
        **dumps_kwargs: Any,
    ) -> dict[str, Any]:
        return json.loads(
            self.json(
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                skip_defaults=skip_defaults,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
                encoder=encoder,
                models_as_dict=models_as_dict,
                **dumps_kwargs,
            )
        )

    class Config:
        json_encoders = {bson.ObjectId: str}
        allow_population_by_field_name = True
        validate_assignment = True


class TimeStampedModel(BaseModel):
    created_at: types.DateTime = field.DateTimeField()
    updated_at: types.DateTime = field.DateTimeField()

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)

        created_at_field = cls.__fields__.pop("created_at")
        updated_at_field = cls.__fields__.pop("updated_at")
        cls.__fields__ |= dict(created_at=created_at_field, updated_at=updated_at_field)
